//*****************************************************************************
// Description     : Main sample program to measure histogram
// Target Device   : APV8508, APV8104
// Target Platform : Linux, C
// Date            : August 2017
// Version         : 1.0
//*****************************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "file.h"
#include "sitcp.h"
#include "device.h"
#include "device_define.h"
#include "local_define.h"
#include "configHist.h"


//
// Local functions declaration
//
static int  config_device(DEVICE_COM_INFO*, unsigned qword);
static int  run_measurement(DEVICE_COM_INFO*, unsigned qword);
static int  receive_data(DEVICE_COM_INFO*);


//#########################################################################
//  Name      : main
//  Function  : entry point
//  Arguments :
//      in argc : number of parameters
//      in argv : parameters
//  Return :
//      0 : success
//      1 : failure
//  Note : 
//#########################################################################
int main(int argc, char **argv)
{
    int run_ready;
    int ret;
    unsigned qword meas_count;
    DEVICE_COM_INFO cominfo;

    run_ready = 0;
    memset(&cominfo, 0, sizeof(cominfo));

    // 1. connect to the device
    ret = device_connect(&cominfo);
    if (ret != 0) {
        printf("device connect error(%d)...\n", ret);
        return(1);
    }

    // 2. set parameters to the device
    meas_count = device_msec2count(MEASUREMENT_TIME);   // convert msec to count
    ret = config_device(&cominfo, meas_count);
    if (ret == 0) {
        run_ready = 1;
    }
    else {
        printf("setting device config error(%d)...\n", ret);
    }

    if (run_ready == 1) {   // ready to run
        // 3. start measurement
        printf("start measurement!!!\n");
        ret = run_measurement(&cominfo, meas_count);

        printf("finish measurement...\n");

        if (ret >= 0) {
            printf("measurement completed!!!\n");
        }
        else {
            printf("measurement error(%d)...\n", ret);
        }
    }

    // 4. post processing
    device_disconnect(&cominfo);

    return(0);
}


//#########################################################################
//  Name      : config_device
//  Function  : set parameters to device
//  Arguments :
//      in pComInfo : pointer to the structure data for referring sockets
//      in ms_count : measurement time (counts)
//  Return :
//       0 : success
//      -n : failed at nth parameter in configHist[]
//      -10001: failed at mode setting
//      -10002: failed at measurement mode setting
//      -10003: failed at measurement time setting
//  Note : 
//#########################################################################
static int config_device(DEVICE_COM_INFO *pComInfo, unsigned qword ms_count)
{
    int i, ret;

    // set by each command
    for (i = 0; configHist[i].cmd_addr != 0; i++) {
        ret = device_write_16bit(pComInfo, configHist[i].cmd_addr, configHist[i].value);
        if (ret < 0)  {
            return(-1 * (i+1));
        }
    }
    // wait for completion of each command processing
    msleep(100);

    // set mode to histogram
    ret = device_write_16bit(pComInfo, DEVICE_COMMAND_MOD, DEVICE_PARAM_MOD_HIST);
    if (ret < 0)  {
        return(-10001);
    }

    // set measurement mode to real time
    ret = device_write_16bit(pComInfo, DEVICE_COMMAND_MMD, DEVICE_PARAM_MMD_REALTIME);
    if (ret < 0)  {
        return(-10002);
    }

    // set measurement time
    ret = device_write_64bit(pComInfo, DEVICE_COMMAND_MTM, ms_count);

    if (ret < 0) {
        return(-10003);
    }

    return(0);
}


//#########################################################################
//  Name      : run_measurement
//  Function  : measurement processing
//  Arguments :
//      in pComInfo : pointer to the structure data for referring sockets
//      in ms_count : measurement time (counts)
//  Return :
//       0 : success
//      -1 : failed while measurement
//      -2 : failed to write file header
//      -3 : failed to receive histogram data
//  Note : 
//#########################################################################
static int run_measurement(DEVICE_COM_INFO *pComInfo, unsigned qword ms_count)
{
    int ret;
    unsigned qword now_count;
    unsigned int now_msec;
    unsigned short status;
    unsigned int output_count[DEVICE_MAX_CHNUM];
    unsigned int output_cps[DEVICE_MAX_CHNUM];

    // clear data
    device_write_3step(pComInfo, DEVICE_COMMAND_CLR);
    // wait for completion of clear processing
    msleep(100);

    // start measurement
    device_start_measurement(pComInfo);

    // display top 2 lines of status field to explain each column
    printf("<time(msec)>   |<output rate(cps)>\n");
    if (DEVICE_MAX_CHNUM > 4) {
        printf(" real/total    |  CH1     CH2     CH3     CH4     CH5     CH6     CH7     CH8\n");
        printf("---------------+-------+-------+-------+-------+-------+-------+-------+-------\n");
    }else{
        printf(" real/total    |  CH1     CH2     CH3     CH4\n");
        printf("---------------+-------+-------+-------+-------\n");
    }

    while(1) {
        // read status
        ret = device_read_status(pComInfo, &status, &now_count, output_count, output_cps);

        if (ret < 0) { // communication error
            // quit measurement
            break;
        }
        // status is 'stopped' or measurement time elapsed
        if ((status == DEVICE_PARAM_AQS_STOP) || (now_count >= ms_count)) {
            break;
        }

        msleep(200);   // wait for 200ms
    }

    // stop measurement
    device_stop_measurement(pComInfo);

    // quit by error
    if (ret < 0) {
        return(-1);
    }

    // write file header
    now_msec  = (unsigned int)(device_count2msec(now_count));  // convert count to msec
    ret =  file_write_header(SAVE_FILENAME, now_msec, output_count, output_cps);
    if (ret < 0) {
        return(-2);
    }

    // receive histogram data then write out to file
    ret = receive_data(pComInfo);
    if (ret < 0) {
        return(-3);
    } else {
        return(0);
    }
}


//#########################################################################
//  Name      : receive_data
//  Function  : receiving collected data
//  Arguments :
//      in pComInfo : pointer to the structure data for referring sockets
//  Return :
//       0 : success
//      -1 : failed to receive
//      -2 : failed to write file
//  Note : 
//#########################################################################
static int receive_data(DEVICE_COM_INFO *pComInfo)
{
    int ret, i;
    int result;
    unsigned int hdata[DEVICE_MAX_CHNUM][NUM_DATA];

    result = 0;

    memset(hdata, 0, sizeof(hdata));
    for (i = 0; i < DEVICE_MAX_CHNUM; i++) {
        // receiving histogram data
        ret = device_read_histogram(pComInfo,
                                   (char*)(&hdata[i]), sizeof(hdata[i]),
                                    DEVICE_TIMEOUT_DATA, i);
    }
    
    if (ret > 0) {
        // write data to file
        ret =  file_write_data(SAVE_FILENAME, hdata);
        if (ret < 0) {
            // failed to write
            result = -2;
        }
    } else {
        // failed to receive data
        result = -1;
    }

    return(result);
}
