//*****************************************************************************
// Description     : Definition of functions for SiTCP communication
// Target Device   : APV8508, APV8104
// Target Platform : Linux, C
// Date            : August 2017
// Version         : 1.0
//*****************************************************************************
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "sitcp.h"
#include "sys_compatible.h"
#include "device_define.h"
#include "command.h"
#include "local_define.h"

//
// Local functions declaration
//
static int sitcp_cmd_init(unsigned short, char*, int, int*);
static int sitcp_data_init(unsigned short, char*, int, int*);
static int sitcp_free(int);
static int sitcp_read(int, unsigned int, int, SITCP_FORMAT*);
static int sitcp_write(int, unsigned int, void*, int);
static int sitcp_recv_cmd(int, char*, int, int);


//#########################################################################
//  Name      : sitcp_connect
//  Function  : connect to the device
//  Arguments :
//      in/out pCom : pointer to the structure data for referring sockets
//  Return :
//       0 : success
//      -1 : failed to connect port for command communication
//      -2 : failed to connect port for fast data transfer
//  Note : 
//#########################################################################
int sitcp_connect(SITCP_COM *pCom)
{
    int sock_cmd, sock_data;
    int ret;

    sock_cmd = 0;
    sock_data = 0;

    // connect to port for command communication	
    ret = sitcp_cmd_init(DEVICE_PORT_CTRL_OWNPC,   // Source port number for command
                         IPADDR_OWNPC,             // IP address of PC
                         SITCP_SOCKBUF_CMD,       // Size of the buffer of socket
                         &sock_cmd);

    if (ret < 0) {  // failed to connect port for command communication
		return(-1);
	}

    // connect to port for fast data transfer	
    ret = sitcp_data_init(DEVICE_PORT_DATARECV,    // Port number for data reception
                          IPADDR_DEVICE,           // IP address of target device
                          SITCP_SOCKBUF_DATA,     // Size of the buffer of socket
                          &sock_data);

    if (ret < 0) {  // failed to connect port for fast data transfer
        // release the port for command communication
        sitcp_free(sock_cmd);
        return(-2);
    }

	// memorize sockets for future reference
	pCom->cmd_socket = sock_cmd;
	pCom->data_socket = sock_data;

    return(0);
}


//#########################################################################
//  Name      : sitcp_disconnect
//  Function  : disconnect from the device
//  Arguments :
//      in pCom : pointer to the structure data for referring sockets
//  Return : None
//  Note : 
//#########################################################################
void sitcp_disconnect(SITCP_COM *pCom)
{
    if (pCom != NULL) {
        sitcp_free(pCom->cmd_socket);
        sitcp_free(pCom->data_socket);
    }
}


//#########################################################################
//  Name      : sitcp_read
//  Function  : command communication(read request)
//  Arguments :
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      in  size     : size of data to be read(byte)
//      out ackdt    : response data
//  Return :
//       0 : success
//      -1 : no response within a period of timeout value
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//  Note : This function doesn't convert received data. It's up to caller.
//#########################################################################
int sitcp_read(int sock_cmd, unsigned int cmd_addr, int size, SITCP_FORMAT *ackdt)
{
    struct sockaddr_in saddr;
    SITCP_FORMAT reqdt, rcvdt;
    SITCP_HEADER cmpdt;
    int result;
    int sbyte, rbyte;
    int ssize, rsize;
    unsigned char wid;
    static unsigned char gid = 0;   // global ID to distinguish response

    result = 0;
    wid = gid++;

    // clear area for reception
    memset(&rcvdt, 0, sizeof(rcvdt));

    // set size of packet for request and response
    ssize = sizeof(SITCP_HEADER);
    rsize = sizeof(SITCP_HEADER) + size;

    // construct packet for read request
    memset(&reqdt, 0, sizeof(reqdt));
    reqdt.header.ver_type = SITCP_VER_TYPE;
    reqdt.header.cmd_flag = SITCP_READ_REQ_CMDFLAG;
    reqdt.header.id       = wid;
    reqdt.header.length   = (unsigned char)(size);
    reqdt.header.address  = htonl(cmd_addr);

    // set target address information
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_port   = htons(DEVICE_PORT_CTRL_DEVICE);     // target port
    saddr.sin_addr.s_addr = inet_addr(IPADDR_DEVICE);      // target IP address

    // send the packet to the target
    sbyte = sendto(sock_cmd, (char*)(&reqdt), ssize, 0,
                  (struct sockaddr*)(&saddr), sizeof(saddr));

    if (sbyte > 0) {
        // receive response
        rbyte = sitcp_recv_cmd(sock_cmd, (char*)(&rcvdt), rsize, DEVICE_TIMEOUT_CMD);

        if (rbyte >= rsize) {
            // construct header part of response packet from request one
            // by change flag, which is only difference, from 'request' to 'acknowledge'
            cmpdt = reqdt.header;
            cmpdt.cmd_flag = SITCP_READ_ACK_CMDFLAG;

            // check if header part is expected one
            if (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) {
                result = -3;   // not a response to the request
            }

            // convert command(address) from network byte order to host byte order
            rcvdt.header.address = ntohl(rcvdt.header.address);
        }
        else { // short of data size
            result = -2;
        }
    }
    else { // receive error
        result = -1;
    }

    // pass response data to caller
    ackdt[0] = rcvdt;

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_16bit
//  Function  : command communication(read request for 16bit value)
//  Arguments :
//      in  pCom     : pointer to the structure data for referring sockets
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned short sitcp_read_16bit(SITCP_COM *pCom, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned short result;

    result = 0;

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(pCom->cmd_socket, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = htons(rdat.data.u16[0]);
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_32bit
//  Function  : command communication(read request for 32bit value)
//  Arguments :
//      in  pCom     : pointer to the structure data for referring sockets
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned int sitcp_read_32bit(SITCP_COM *pCom, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned int result;

    result = 0;

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(pCom->cmd_socket, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = htonl(rdat.data.u32[0]);
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_48bit
//  Function  : command communication(read request for 48bit value)
//  Arguments :
//      in  pCom     : pointer to the structure data for referring sockets
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned qword sitcp_read_48bit(SITCP_COM *pCom, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    char abyte[8];
    unsigned qword result;

    result = (unsigned qword)(0);

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(pCom->cmd_socket, cmd_addr, 6, &rdat);

    if (ret == 0) {
        *(short*)(&abyte[0]) = 0x00;   // clear top 2bytes
        memcpy(&abyte[2], &rdat.data.u08[0], 6);
        result = sitcp_ntoh64(*(unsigned qword*)(&abyte[0]));
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return  result;
}


//#########################################################################
//  Name      : sitcp_read_64bit
//  Function  : command communication(read request for 64bit value)
//  Arguments :
//      in  pCom     : pointer to the structure data for referring sockets
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned qword sitcp_read_64bit(SITCP_COM *pCom, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned qword result;

    result = (unsigned qword)(0);

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(pCom->cmd_socket, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = sitcp_ntoh64(rdat.data.u64[0]);
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_write
//  Function  : command communication(write request)
//  Arguments :
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      in  wbuffer  : data to be written 
//      in  size     : size of data to be written(byte)
//  Return :
//       0 : success
//      -1 : no response within a period of timeout value
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//  Note : This function doesn't convert written data. It's up to caller.
//#########################################################################
int sitcp_write(int sock_cmd, unsigned int cmd_addr, void *wbuffer, int size)
{
    struct sockaddr_in saddr;
    SITCP_FORMAT reqdt, rcvdt;
    SITCP_HEADER cmpdt;
    int result;
    int sbyte, rbyte;
    int ssize, rsize;
    unsigned char wid;
    static unsigned char gid = 0;   // global ID to distinguish response

    result = 0;
    wid = gid++;

    // clear area for reception
    memset(&rcvdt, 0, sizeof(rcvdt));

    // set size of packet for request and response
    ssize = sizeof(SITCP_HEADER) + size;
    rsize = ssize;

    // construct packet for write request
    memset(&reqdt, 0, sizeof(reqdt));
    reqdt.header.ver_type = SITCP_VER_TYPE;
    reqdt.header.cmd_flag = SITCP_WRITE_REQ_CMDFLAG;
    reqdt.header.id       = wid;
    reqdt.header.length   = (unsigned char)(size);
    reqdt.header.address  = htonl(cmd_addr);
    memcpy(&reqdt.data, wbuffer, size);

    // set target address information
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_port   = htons(DEVICE_PORT_CTRL_DEVICE);     // target port
    saddr.sin_addr.s_addr = inet_addr(IPADDR_DEVICE);      // target IP address

    // send the packet to the target
    sbyte = sendto(sock_cmd, (char*)(&reqdt), ssize, 0,
                  (struct sockaddr*)(&saddr), sizeof(saddr));

    if (sbyte > 0) {
        // receive response
        rbyte = sitcp_recv_cmd(sock_cmd, (char*)(&rcvdt), rsize, DEVICE_TIMEOUT_CMD);

        if (rbyte >= rsize) {
            // construct header part of response packet from request one
            // by change flag, which is only difference, from 'request' to 'acknowledge'
            cmpdt = reqdt.header;
            cmpdt.cmd_flag = SITCP_WRITE_ACK_CMDFLAG;

            // check if header part is expected one
            if (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) {
                result = -3;   // not a response to the request
            }

            // convert command(address) from network byte order to host byte order
            rcvdt.header.address = ntohl(rcvdt.header.address);
        }
        else { // short of data size
            result = -2;
        }
    }
    else { // receive error
        result = -1;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_16bit
//  Function  : command communication(write request for 16bit value)
//  Arguments :
//      in pCom     : pointer to the structure data for referring sockets
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_16bit(SITCP_COM *pCom, unsigned int cmd_addr, unsigned short wdata)
{
    int result;
    unsigned short value;

    value = htons(wdata);
    result = sitcp_write(pCom->cmd_socket, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_32bit
//  Function  : command communication(write request for 32bit value)
//  Arguments :
//      in pCom     : pointer to the structure data for referring sockets
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_32bit(SITCP_COM *pCom, unsigned int cmd_addr, unsigned int wdata)
{
    int result;
    unsigned int value;

    value = htonl(wdata);
    result = sitcp_write(pCom->cmd_socket, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_48bit
//  Function  : command communication(write request for 48bit value)
//  Arguments :
//      in pCom     : pointer to the structure data for referring sockets
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_48bit(SITCP_COM *pCom, unsigned int cmd_addr, unsigned qword wdata)
{
    int result;
    unsigned char *p, *q;
    unsigned qword value, wvalue;

    value  = 0;
    wvalue = sitcp_hton64(wdata);

    p = (unsigned char*)(&value);
    q = (unsigned char*)(&wvalue);
    memcpy(&p[0], &q[2], 6);   // copy valid 48bit to upper bit of destination area
    result = sitcp_write(pCom->cmd_socket, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_64bit
//  Function  : command communication(write request for 64bit value)
//  Arguments :
//      in pCom     : pointer to the structure data for referring sockets
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_64bit(SITCP_COM *pCom, unsigned int cmd_addr, unsigned qword wdata)
{
    int result;
    unsigned qword value;

    value  = sitcp_hton64(wdata);
    result = sitcp_write(pCom->cmd_socket, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_recv_cmd
//  Function  : command communication(receive response)
//  Arguments :
//      in  sock    : file descriptor for command communication
//      in  rbuffer : buffer for received data
//      in  size    : size of data to be received(byte)
//      in  timeout : receive timeout value(msec)
//  Return : 
//      positive value: received data size
//      -11 : no response within a period of timeout value
//      -12 : no data
//  Note :
//#########################################################################
int sitcp_recv_cmd(int sock, char *rbuffer, int size, int timeout)
{
    struct sockaddr_in saddr;
    struct timeval tval;
    socklen_t sock_len;
    fd_set fdsock;
    int ret;
    int result;
    int recv_size;

    result = 0;

    // set timeout value
    memset(&tval, 0, sizeof(tval));
    tval.tv_sec  = (int)(timeout / 1000);
    tval.tv_usec = (int)((timeout % 1000) * 1000);

    // wait until received data is ready
    FD_ZERO(&fdsock);
    FD_SET(sock, &fdsock);
    ret = select((sock + 1), &fdsock, NULL, NULL, &tval);

    if (ret != 0) {
        if (FD_ISSET(sock, &fdsock) != 0) {
            // receive message
            sock_len = sizeof(saddr);
            memset(&saddr, 0, sizeof(saddr));
            recv_size = recvfrom(sock, rbuffer, size, 0, (struct sockaddr*)(&saddr), &sock_len);
            result = (int)(recv_size);     // set received data size
        }
        else { // no data
            result = -12;
        }
    }
    else { // timeout
        result = -11;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_recv_data
//  Function  : receive data by fast data transfer(TCP)
//  Arguments :
//      in  pCom    : pointer to the structure data for referring sockets
//      in  rbuffer : buffer for received data
//      in  size    : size of data to be received(byte)
//      in  timeout : receive timeout value(msec)
//  Return : 
//      positive value: received data size
//      -1 : communication error
//  Note :
//#########################################################################
int sitcp_recv_data(SITCP_COM *pCom, char *rbuffer, int size, int timeout)
{
    struct timeval tval, tval2nd;
    fd_set fdsock;
    int ret;
    int result;
    int rbyte, rsize;
    unsigned qword count;
	int sock = pCom->data_socket;

    result = 0;
    count = 0;

    // save timeout value for 1st read
    memset(&tval, 0, sizeof(tval));
    tval.tv_sec  = (int)(timeout / 1000);
    tval.tv_usec = (int)((timeout % 1000) * 1000);

    // save timeout value for 2nd read or later
    memset(&tval2nd, 0, sizeof(tval2nd));
    tval2nd.tv_sec  = (int)(timeout / 10);
    tval2nd.tv_usec = (int)((timeout % 10) * 10);

    FD_ZERO(&fdsock);
    FD_SET(sock, &fdsock);

    while(1) {
        if (count <= 1 && timeout > 0) {
            // wait until received data is ready
            ret = select((sock + 1), &fdsock, 0, 0, &tval);

            if (ret <= 0) {
                // quit by timeout
                break;
            }
            if (FD_ISSET(sock, &fdsock) == 0) {
                // quit by no data
                break;
            }

            // update timeout value for 2nd or later read
            tval = tval2nd;
        }

        // check size of received message
        rbyte = recv(sock, (char*)(&rbuffer[result]), 1, MSG_PEEK | MSG_DONTWAIT);
        if (rbyte <= 0) {
            break;
        }

        // update request data size
        rsize = size - result;

        // receive message
        rbyte = recv(sock, (char*)(&rbuffer[result]), rsize, MSG_DONTWAIT);
        if (rbyte > 0) {
            // update total received data size
            result += rbyte;

            if (result >= size) {
                // finish when total received data size reached to request one
                break;
            }
        }
        else {  // failed to receive
            result = -1;
            break;
        }

        count++;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_cmd_init
//  Function  : connect to device for command communication(UDP)
//  Arguments :
//      in  portn     : port number for receiving response
//      in  ipaddr    : IP address of PC; Normally leave 0.0.0.0
//      in  sock_size : maximum socket receive buffer size(bytes)
//      out psock     : pointer to the file descriptor for the socket
//  Return :
//       0 : success
//      -1 : failed to create socket
//      -2 : bind error
//  Note : 
//#########################################################################
int sitcp_cmd_init(unsigned short portn, char *ipaddr, int sock_size, int *psock)
{
    struct sockaddr_in saddr;
    int sock;
    int ret;
    int result;
    int optval;
    socklen_t socklen;

    result = 0;
    *psock = 0;

    // create UDP socket
    sock = socket(AF_INET, SOCK_DGRAM, 0);

    if (sock > 0) { // socket created
        if (sock_size > 0) {
            // change maximum socket receive buffer size if required
            optval  = sock_size;
            socklen = sizeof(optval);
            ret = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &optval, socklen);
        }

        // assign own address to the socket
        memset(&saddr, 0, sizeof(saddr));
        saddr.sin_family = AF_INET;
        saddr.sin_port   = htons(portn);            // port number for receive
        saddr.sin_addr.s_addr = inet_addr(ipaddr);  // IP address of PC
        ret = bind(sock, (struct sockaddr*)(&saddr), sizeof(saddr));

        if (ret == 0) { // bind succeeded
            *psock = sock;
            result = 1;
        }
        else {  // bind error
            close(sock);
            result = -2;
        }
    }
    else {  // failed to create socket
        result = -1;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_data_init
//  Function  : connect to device for fast data transfer(TCP)
//  Arguments :
//      in  portn     : port number for receiving data
//      in  ipaddr    : IP address of target device
//      in  sock_size : maximum socket receive buffer size(bytes)
//      out psock     : pointer to the file descriptor for the socket
//  Return :
//       0 : success
//      -1 : failed to create socket
//      -2 : connect error
//  Note : 
//#########################################################################
int sitcp_data_init(unsigned short portn, char *ipaddr, int sock_size, int *psock)
{
    struct sockaddr_in saddr;
    int sock;
    int ret;
    int result;
    int optval;
    socklen_t socklen;

    result = 0;
    *psock = 0;

    // create TCP socket
    sock = socket(AF_INET, SOCK_STREAM, 0);

    if (sock > 0) { // socket created
        if (sock_size > 0) {
            // change maximum socket receive buffer size if required
            optval  = sock_size;
            socklen = sizeof(optval);
            ret = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &optval, socklen);
        }

        // connect to target device
        memset(&saddr, 0, sizeof(saddr));
        saddr.sin_family = AF_INET;
        saddr.sin_port   = htons(portn);            // port number for receiving data
        saddr.sin_addr.s_addr = inet_addr(ipaddr);    // IP address of target device
        ret = connect(sock, (struct sockaddr*)(&saddr), sizeof(saddr));

        if (ret == 0) { // connect succeeded
            *psock = sock;
            result = 1;
        }
        else {  // connect error
            close(sock);
            result = -2;
        }
    }
    else {  // failed to create socket
        result = -1;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_free
//  Function  : close socket
//  Arguments :
//      in sock : file descriptor for the socket
//  Return :
//        0 : success
//       -1 : invalid descriptor
//  Note : 
//#########################################################################
int sitcp_free(int sock)
{
    int result;

    if (sock > 0) {
        close(sock);
        result = 0;
    } else {
        result = -1;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_ntoh64
//  Function  : convert 64bit value from network byte order to host byte order 
//  Arguments :
//      in value : 64bit value in network byte order
//  Return : 64bit value in host byte order
//  Note : 
//#########################################################################
unsigned qword sitcp_ntoh64(unsigned qword value)
{
    return sitcp_hton64(value);
}


//#########################################################################
//  Name      : sitcp_hton64
//  Function  : convert 64bit value from host byte order to network byte order
//  Arguments :
//      in value : 64bit value in host byte order
//  Return : 64bit value in network byte order
//  Note : 
//#########################################################################
unsigned qword sitcp_hton64(unsigned qword value)
{
    unsigned char *p, *q;
    unsigned qword result;
    const static char adefs[2] = { 0x12, 0x34 };  // for identifying endian

    result = 0;

    if (*(short*)(adefs) != 0x1234) {    // little endian
        // invert byte order
        p = (unsigned char*)(&result);
        q = (unsigned char*)(&value);

        p[0] = q[7];
        p[1] = q[6];
        p[2] = q[5];
        p[3] = q[4];
        p[4] = q[3];
        p[5] = q[2];
        p[6] = q[1];
        p[7] = q[0];
    }
    else {  // big endian; no need to convert
        result = value;
    }

    return(result);
}
